package redner

import (
	"context"
	"database/sql"
	"errors"
	"fmt"
	"net/smtp"
	"time"

	"github.com/Masterminds/squirrel"
	"github.com/jmoiron/sqlx"
	"github.com/lib/pq"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
)

var ErrUnknownSMTPAuthType = errors.New("unknown smtp-auth type")

//nolint:lll
type MailSender struct {
	Mode         string `long:"mode" ini-name:"mode" default:"smtp" choice:"smtp"`
	SMTPFrom     string `long:"smtp-from" ini-name:"smtp-from" description:"email sender address"`
	SMTPHost     string `long:"smtp-host" ini-name:"smtp-host" description:"smtp server"`
	SMTPPort     string `long:"smtp-port" ini-name:"smtp-port" description:"smtp port"`
	SMTPAuth     string `long:"smtp-auth" ini-name:"smtp-auth" default:"CRAMMD5" choice:"CRAMMD5" choice:"PLAIN"`
	SMTPUser     string `long:"smtp-user" ini-name:"smtp-user"`
	SMTPPassword string `long:"smtp-password" ini-name:"smtp-password"`

	SMTPMaxAttempts int `long:"smtp-max-attempts" ini-name:"smtp-max-attempts" default:"10" description:"How many times to attempt sending before considering it a failure"`

	addr string
	auth smtp.Auth
}

func (s MailSender) Start(ctx context.Context, dsn string, db *sqlx.DB) error {
	log := zerolog.Ctx(ctx)
	log.Info().Msg("MailSender is starting")

	switch s.SMTPAuth {
	case "CRAMMD5":
		log.Info().Msg("mail: Setting up CRAMMD5 smtp auth")

		s.auth = smtp.CRAMMD5Auth(s.SMTPUser, s.SMTPPassword)
	case "PLAIN":
		log.Info().Msg("mail: Setting up PLAIN smtp auth")

		s.auth = smtp.PlainAuth("", s.SMTPUser, s.SMTPPassword, s.SMTPHost)
	default:
		return fmt.Errorf("%w: %s", ErrUnknownSMTPAuthType, s.SMTPAuth)
	}

	s.addr = s.SMTPHost + ":" + s.SMTPPort

	messageAvailable, signalMessageAvailable := NewCondChan(func() bool {
		return ctx.Err() != nil
	})
	go func() {
		<-ctx.Done()
		signalMessageAvailable()
	}()

	listener := pq.NewListener(dsn, time.Second, time.Minute,
		func(ev pq.ListenerEventType, err error) {
		})

	if err := listener.Listen("email_queue_notify"); err != nil {
		return err
	}

	go func() {
		for {
			select {
			case <-ctx.Done():
				return
			case <-listener.Notify:
				signalMessageAvailable()
			}
		}
	}()

	go s.run(ctx, db, messageAvailable)

	log.Info().Msg("MailSender is ready")

	return nil
}

func (s MailSender) run(ctx context.Context, db *sqlx.DB, messageAvailable <-chan struct{}) {
	log := zerolog.Ctx(ctx)

	ticker := time.NewTicker(time.Minute)
	defer ticker.Stop()

	s.flush(ctx, db)

	log.Info().Msg("now waiting for emails to send")

	for {
		select {
		case <-ticker.C:
			s.flush(ctx, db)
		case <-messageAvailable:
			s.flush(ctx, db)
		case <-ctx.Done():
			return
		}
	}
}

func (s MailSender) flush(ctx context.Context, db *sqlx.DB) {
	log := zerolog.Ctx(ctx)

	for {
		if ctx.Err() != nil {
			log.Debug().Msg("mailsender: context done, stopping")

			return
		}

		if err := s.handle(ctx, db); err != nil {
			if errors.Is(err, sql.ErrNoRows) {
				log.Debug().Msg("no mail in the queue")

				return
			}

			// handle the rest of the queue
			log.Err(err).Msg("could not send mail, waiting a little")
		}
	}
}

func (s MailSender) handle(ctx context.Context, sqlDB *sqlx.DB) error {
	log := zerolog.Ctx(ctx)
	log.Debug().Msg("MailSender.handle")

	tx, err := database.Begin(ctx, sqlDB)
	if err != nil {
		return err
	}

	defer tx.RollbackIfOpened(*log)

	db := database.NewSQLHelper(ctx, tx, *log)

	var email models.Email
	if err := db.Get(&email,
		squirrel.Select(models.EmailColumns...).
			From(models.EmailTable).
			Where(squirrel.And{
				squirrel.Eq{"status": []string{"pending", "error"}},
				squirrel.LtOrEq{"reattempt_at": time.Now().UTC()},
			}).
			OrderBy(models.EmailIDColumn).
			Limit(1).
			Suffix("FOR UPDATE SKIP LOCKED"),
	); err != nil {
		return err
	}

	var to []string
	if err := email.To.AssignTo(&to); err != nil {
		panic(err)
	}

	if err := smtp.SendMail(s.addr, s.auth, s.SMTPFrom, to, email.Msg); err != nil {
		email.StatusChanged = time.Now().UTC()
		email.Status = "error"
		email.LastError = err.Error()
		email.Attempts++

		if email.Attempts >= s.SMTPMaxAttempts {
			email.Status = "failed"
		} else {
			delay := time.Minute * time.Duration(email.Attempts)
			if delay > time.Hour {
				delay = time.Hour
			}

			email.ReattemptAt = time.Now().Add(delay).UTC()
		}

		if err := db.Update(&email); err != nil {
			return err
		}

		if err := tx.Commit(); err != nil {
			return err
		}

		return err
	}

	email.StatusChanged = time.Now().UTC()
	email.Status = "sent"
	email.LastError = ""

	if err := db.Update(&email); err != nil {
		return err
	}

	if err := tx.Commit(); err != nil {
		return err
	}

	log.Debug().Strs("to", to).Msg("Successfully sent an email")

	return nil
}
