package redner

import (
	"context"
	"database/sql"
	"errors"
	"fmt"
	"sync"
	"time"

	"github.com/jmoiron/sqlx"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/rendering"
	"orus.io/orus-io/rednerd/utils"
)

const LogRenderBatchSize = 1000

type contextKey int

var (
	userNameContextKey contextKey = 1
	ErrInvalidInput               = errors.New("invalid input")

	renderPipelineDuration = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "rednerd_render_pipeline",
			Help:       "Render pipeline duration and count",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
		},
		[]string{
			"template_language",
			"from_type",
			"to_type",
			"status",
		},
	)
)

func NewRenderer(
	ctx context.Context,
	db *sqlx.DB,
	registry *rendering.EngineRegistry,
	templateRegistry *rendering.TemplateEngineRegistry,
	log zerolog.Logger,
) *Renderer {
	r := Renderer{
		db:               db,
		registry:         registry,
		templateRegistry: templateRegistry,
		logRender:        make(chan *models.RenderLog, LogRenderBatchSize*2),
	}
	ctx, r.stop = context.WithCancel(ctx)
	r.wg.Add(1)

	go r.run(ctx, db, log)

	return &r
}

type DocumentCallback func(context.Context, *database.SQLHelper, *models.Document) error

type Renderer struct {
	db               *sqlx.DB
	registry         *rendering.EngineRegistry
	templateRegistry *rendering.TemplateEngineRegistry
	logRender        chan *models.RenderLog
	wg               sync.WaitGroup
	stop             func()

	onDocRender []DocumentCallback
}

func (r *Renderer) OnDocRender(cb DocumentCallback) {
	r.onDocRender = append(r.onDocRender, cb)
}

func (r *Renderer) run(ctx context.Context, dbc *sqlx.DB, log zerolog.Logger) {
	defer r.wg.Done()

	db := database.NewSQLHelper(ctx, dbc, log)

	timer := time.NewTimer(0)
	defer timer.Stop()

	for {
		select {
		case rlog := <-r.logRender:
			if rlog == nil {
				continue
			}

			if rlog.ID != 0 {
				if err := db.Update(rlog); err != nil {
					log.Err(err).Msg("could not update render log")
				}
			} else {
				if _, err := db.InsertNoPKey(rlog); err != nil {
					log.Err(err).Msg("could not insert render log")
				}
			}
		case <-timer.C:
		case <-ctx.Done():
			return
		}
	}
}

func (r *Renderer) Close() {
	close(r.logRender)
	r.stop()
	r.wg.Wait()
}

func (r *Renderer) Render(
	ctx context.Context,
	username string,
	template *models.Template,
	data models.Dataset,
	document *models.Document,
	metadata models.Metadata,
	toType string,
	enableRequestLogging bool,
) ([]*models.Document, error) {
	var (
		log    = zerolog.Ctx(ctx)
		result []*models.Document

		templateLanguage string
		fromType         string
	)

	pipelineTimer := utils.NewVecTimer(renderPipelineDuration)

	db := database.NewSQLHelper(ctx, r.db, *log)

	rlog := models.NewRenderLog(
		pipelineTimer.GetBegin(),
		fromType,
		toType,
		username,
	)

	if template != nil && template.Account != "" && template.Name != "" {
		templatedb, err := LoadTemplate(&db, template.Account, template.Name)
		if err != nil {
			if errors.Is(err, sql.ErrNoRows) {
				return nil, fmt.Errorf(
					"%w: no such template '%s/%s'",
					ErrInvalidInput, template.Account, template.Name)
			}

			return nil, err
		}

		templatedb.FromDB()
		template = &templatedb.Template
		rlog.SetTemplate(templatedb.ID, templatedb.InternalVersion, template.Language)
	}

	if enableRequestLogging {
		rlog.SetDebugQuery(&models.RenderRequest{
			Accept:   models.Accept(toType),
			Data:     data,
			Metadata: metadata,
			Document: document,
			Template: template,
		})

		query := database.SQLInsertNoPKey(rlog)
		query = query.Suffix("RETURNING id")

		if err := db.Get(&rlog.ID, query); err != nil {
			log.Err(err).Msg("could not save the render log entry")
		}
	}

	if template != nil {
		templateLanguage = template.Language
		fromType = template.Produces

		pipeline, err := r.templateRegistry.NewPipeline(
			ctx, r.registry, template, metadata, toType,
		)
		if err != nil {
			if errors.Is(err, ErrInvalidInput) {
				return nil, models.NewRenderFailedError(err, nil)
			}

			if errors.Is(err, rendering.ErrUnknownTemplateLanguage) {
				return nil, fmt.Errorf("%w: unknown template language: %s", ErrInvalidInput, template.Language)
			}

			return nil, fmt.Errorf("%w: error loading rendering pipeline: %w", ErrInvalidInput, err)
		}

		for _, record := range data {
			var doc models.Document

			result = append(result, &doc)

			w := rendering.NewDocumentWriter(&doc)
			if err := pipeline.Render(ctx, w, record); err != nil {
				pipelineTimer.ObserveDurationWithLabelValues(
					templateLanguage, fromType, toType, "failure")

				if errors.Is(err, ErrInvalidInput) {
					return nil, models.NewRenderFailedError(err, result)
				}

				return nil, err
			}

			for _, cb := range r.onDocRender {
				if err := cb(ctx, &db, &doc); err != nil {
					doc.AppendRenderErrors([]models.RenderError{
						{Message: err.Error()},
					})
				}
			}
		}
	}

	if document != nil {
		pipeline, err := r.registry.GetPipeline(document.Type, toType)
		if err != nil {
			return nil, fmt.Errorf("%w: %w", ErrInvalidInput, err)
		}

		document.UpdateMetadata(metadata)

		if len(pipeline) != 0 {
			var output models.Document

			result = append(result, &output)

			reader := rendering.NewDocumentReader(document)
			writer := rendering.NewDocumentWriter(&output)

			userContext := SetUsername(ctx, username)
			if err := pipeline.Render(userContext, writer, reader); err != nil {
				if errors.Is(err, ErrInvalidInput) {
					return nil, models.NewRenderFailedError(err, result)
				}

				return nil, err
			}

			for _, cb := range r.onDocRender {
				if err := cb(ctx, &db, &output); err != nil {
					output.AppendRenderErrors([]models.RenderError{
						{Message: err.Error()},
					})
				}
			}
		} else {
			for _, cb := range r.onDocRender {
				if err := cb(ctx, &db, document); err != nil {
					document.AppendRenderErrors([]models.RenderError{
						{Message: err.Error()},
					})
				}
			}

			result = append(result, document)
		}
	}

	rlog.Duration = int(pipelineTimer.ObserveDurationWithLabelValues(
		templateLanguage, fromType, toType, "success") / time.Millisecond)

	r.logRender <- rlog

	return result, nil
}

// GetUsername allows to extract the username from a context
// this is intended to be used by engines to extract the username
// info.
func GetUsername(ctx context.Context) string {
	v := ctx.Value(userNameContextKey)
	if v == nil {
		return ""
	}

	res, ok := v.(string)
	if !ok {
		panic("could not type assert on the userName from context")
	}

	return res
}

// SetUsername set the given username in the given context and returns
// a derived context.
func SetUsername(ctx context.Context, username string) context.Context {
	return context.WithValue(ctx, userNameContextKey, username)
}

func init() {
	prometheus.MustRegister(renderPipelineDuration)
}
